import unittest
import numpy as np
from reval.param_selection import SCParamSelection, ParamSelection
from reval.param_selection import _return_best, _return_knownbest, _create_result_table
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.cluster import AgglomerativeClustering, SpectralClustering
from hdbscan import HDBSCAN
from sklearn.datasets import make_blobs

# Modify to test other functions and parameters
RNDLABELS_ITER = 10
CLASSIFIER = [KNeighborsClassifier(n_neighbors=5),
              LogisticRegression(solver='liblinear'),
              RandomForestClassifier(),
              SVC()]
CLUSTERING = [AgglomerativeClustering(), HDBSCAN(min_samples=10, min_cluster_size=200)]
CLUSTERING_MOD = [AgglomerativeClustering(), HDBSCAN()]
NCLUST_RANGE = list(range(2, 4, 1))
PROCESSES = 1
NFOLD = 2


class TestSCParamSelection(unittest.TestCase):

    @classmethod
    def setUp(cls):
        cls.sc_params = {'s': CLASSIFIER, 'c': CLUSTERING}
        cls.sc_hdbscan = {'s': CLASSIFIER, 'c': [CLUSTERING[-1]]}
        cls.sc_hdbscanmod = {'s': CLASSIFIER, 'c': [CLUSTERING_MOD[-1]]}
        cls.nrand = RNDLABELS_ITER
        cls.cv = NFOLD
        cls.clust_range = NCLUST_RANGE
        cls.processes = PROCESSES
        cls.iter_cv = 1
        cls.scparamselect = SCParamSelection(sc_params=cls.sc_params,
                                             cv=cls.cv,
                                             nrand=cls.nrand,
                                             clust_range=cls.clust_range,
                                             n_jobs=cls.processes,
                                             iter_cv=cls.iter_cv)
        cls.scparamselect_noclrange = SCParamSelection(sc_params=cls.sc_hdbscan,
                                                       cv=cls.cv,
                                                       nrand=cls.nrand,
                                                       n_jobs=cls.processes,
                                                       clust_range=None,
                                                       iter_cv=cls.iter_cv)
        cls.scparamselect_range = SCParamSelection(sc_params=cls.sc_hdbscanmod,
                                                   cv=cls.cv,
                                                   nrand=cls.nrand,
                                                   n_jobs=cls.processes,
                                                   clust_range=cls.clust_range,
                                                   iter_cv=cls.iter_cv)

    def test_fit(self):
        """
        HDBSCAN fails to detect clusters.
        """
        data = np.array([[0] * 20,
                         [1] * 20] * 20)
        self.scparamselect.fit(data, 2)
        self.assertIsInstance(self.scparamselect.best_param_, list)
        self.assertIsInstance(self.scparamselect.best_index_, list)
        self.assertIsInstance(self.scparamselect.cv_results_, dict)
        self.assertSequenceEqual([(self.scparamselect.cv_results_["best_nclust"][bidx],
                                   self.scparamselect.cv_results_["validation_meanerror"][bidx])
                                  for bidx in self.scparamselect.best_index_], [(2, (0.0, (0.0, 0.0)))] *
                                 len(CLASSIFIER) * (len(CLUSTERING) - 1))

    def test_fit_hdbscan(self):
        """
        Only HDBSCAN is selected as clustering.
        """
        data = np.array([[0] * 20,
                         [1] * 20] * 20)
        self.scparamselect_noclrange.fit(data, 2)
        self.assertIsInstance(self.scparamselect_noclrange.cv_results_, dict)
        self.assertSequenceEqual([(self.scparamselect_noclrange.cv_results_["best_nclust"],
                                   self.scparamselect_noclrange.cv_results_["validation_meanerror"])],
                                 [([None] *
                                   len(CLASSIFIER), [None] *
                                   len(CLASSIFIER))])

    def test_fit_hdbscan_mod(self):
        """
        Only HDBSCAN is selected as clustering.
        """
        data = np.array([[0] * 20,
                         [1] * 20] * 20)
        self.scparamselect_range.fit(data, 2)
        self.assertIsInstance(self.scparamselect_range.cv_results_, dict)
        self.assertSequenceEqual([(self.scparamselect_range.cv_results_["best_nclust"][bidx],
                                   self.scparamselect_range.cv_results_["validation_meanerror"][bidx])
                                  for bidx in self.scparamselect_range.best_index_], [(2, (0.0, (0.0, 0.0)))] *
                                 len(CLASSIFIER))

    def test_run_gridsearchcv(self):
        data = np.array([[0] * 20,
                         [1] * 20] * 20)
        perf = self.scparamselect._run_gridsearchcv(data, {'s': CLASSIFIER[0],
                                                           'c': CLUSTERING[0]})
        self.assertEqual(perf, [('s', CLASSIFIER[0]), ('c', CLUSTERING[0]), ('best_nclust', 2),
                                ('mean_train_score', 0.0),
                                ('sd_train_score', 0.0),
                                ('mean_val_score', 0.0),
                                ('sd_val_score', 0.0),
                                ('validation_meanerror', (0.0, (0.0, 0.0))),
                                ('tr_label', None)])

    def test_fit_noclass(self):
        noise = make_blobs(100, 2, centers=1, random_state=42)[0]
        perf = self.scparamselect_noclrange._run_gridsearchcv(noise, {'s': CLASSIFIER[0],
                                                                      'c': CLUSTERING[-1]})
        self.assertIsInstance(perf, list)
        perf = self.scparamselect_range._run_gridsearchcv(noise, {'s': CLASSIFIER[0],
                                                                  'c': CLUSTERING[-1]})
        self.assertIsInstance(perf, list)


class TestParamSelection(unittest.TestCase):

    @classmethod
    def setUp(cls):
        cls.params = {'s': {'n_neighbors': [5, 10, 15]},
                      'c': {'affinity': ['euclidean', 'l1', 'l2', 'manhattan'],
                            'linkage': ['ward', 'complete', 'average', 'single']}}
        cls.params_bis = {'s': {'n_neighbors': [5, 10, 15]},
                          'c': {'min_samples': [10, None],
                                'min_cluster_size': [5, 200]}}
        cls.params_nocomb = {'s': {'n_neighbors': [5, 10, 15]},
                             'c': {'min_samples': [10, None],
                                   'min_cluster_size': [200]}}
        cls.s = KNeighborsClassifier()
        cls.c = AgglomerativeClustering()
        cls.c_bis = HDBSCAN()
        cls.nrand = RNDLABELS_ITER
        cls.cv = NFOLD
        cls.clust_range = NCLUST_RANGE
        cls.processes = PROCESSES
        cls.paramselect = ParamSelection(params=cls.params,
                                         cv=cls.cv,
                                         s=cls.s,
                                         c=cls.c,
                                         nrand=cls.nrand,
                                         n_jobs=cls.processes,
                                         iter_cv=1,
                                         clust_range=cls.clust_range)
        cls.paramselect_bis = ParamSelection(params=cls.params_bis,
                                             cv=cls.cv,
                                             s=cls.s,
                                             c=cls.c_bis,
                                             nrand=cls.nrand,
                                             n_jobs=cls.processes,
                                             iter_cv=1)
        cls.paramselect_nocomb = ParamSelection(params=cls.params_nocomb,
                                                cv=cls.cv,
                                                s=cls.s,
                                                c=cls.c_bis,
                                                nrand=cls.nrand,
                                                n_jobs=cls.processes,
                                                iter_cv=1)

    def test_fit(self):
        # 48 possible parameter combinations - 9 not admissible for
        # agglomerative clustering
        data = np.array([[0] * 20,
                         [1] * 20] * 20)
        self.paramselect.fit(data, 2)
        self.assertIsInstance(self.paramselect.best_param_, list)
        self.assertIsInstance(self.paramselect.best_index_, list)
        self.assertIsInstance(self.paramselect.cv_results_, dict)
        self.assertSequenceEqual([(self.paramselect.cv_results_["best_nclust"][bidx],
                                   self.paramselect.cv_results_["validation_meanerror"][bidx])
                                  for bidx in self.paramselect.best_index_], [(2, (0.0, (0.0, 0.0)))] *
                                 39)

    def test_bis_fit(self):
        data = np.array([[0] * 20,
                         [1] * 20] * 20)
        self.paramselect_bis.fit(data, 2)
        self.assertIsInstance(self.paramselect_bis.best_param_, list)
        self.assertIsInstance(self.paramselect_bis.best_index_, list)
        self.assertIsInstance(self.paramselect_bis.cv_results_, dict)
        self.assertSequenceEqual([(self.paramselect_bis.cv_results_["best_nclust"][bidx],
                                   self.paramselect_bis.cv_results_["validation_meanerror"][bidx])
                                  for bidx in self.paramselect_bis.best_index_], [(2, (0.0, (0.0, 0.0)))] *
                                 6)

    def test_nocomb_fit(self):
        data = np.array([[0] * 20,
                         [1] * 20] * 20)
        self.paramselect_nocomb.fit(data, 2)
        self.assertIsInstance(self.paramselect_nocomb.cv_results_, dict)
        self.assertSequenceEqual([(self.paramselect_nocomb.cv_results_["best_nclust"],
                                   self.paramselect_nocomb.cv_results_["validation_meanerror"])
                                  ], [([None] * 6, [None] * 6)])

    def test_run_gridsearchcv(self):
        data = np.array([[0] * 20,
                         [1] * 20] * 20)
        perf = self.paramselect._run_gridsearchcv(data, {'n_neighbors': 5},
                                                  {'affinity': 'euclidean',
                                                   'linkage': 'ward'})
        self.assertEqual(perf, [('n_neighbors', 5), ('affinity', 'euclidean'),
                                ('linkage', 'ward'),
                                ('best_nclust', 2),
                                ('mean_train_score', 0.0),
                                ('sd_train_score', 0.0),
                                ('mean_val_score', 0.0),
                                ('sd_val_score', 0.0),
                                ('validation_meanerror', (0.0, (0.0, 0.0))),
                                ('tr_label', None)])

        perf_bis = self.paramselect_bis._run_gridsearchcv(data, {'n_neighbors': 5},
                                                          {'min_samples': None,
                                                           'min_cluster_size': 5})
        self.assertEqual(perf_bis, [('n_neighbors', 5), ('min_samples', None),
                                    ('min_cluster_size', 5),
                                    ('best_nclust', 2),
                                    ('mean_train_score', 0.0),
                                    ('sd_train_score', 0.0),
                                    ('mean_val_score', 0.0),
                                    ('sd_val_score', 0.0),
                                    ('validation_meanerror', (0.0, (0.0, 0.0))),
                                    ('tr_label', [1, 0] * 20)])

        perf_nocomb = self.paramselect_bis._run_gridsearchcv(data, {'n_neighbors': 5},
                                                             {'min_samples': None,
                                                              'min_cluster_size': 200})
        self.assertEqual(perf_nocomb, [('n_neighbors', 5), ('min_samples', None),
                                       ('min_cluster_size', 200),
                                       ('best_nclust', None),
                                       ('mean_train_score', None),
                                       ('sd_train_score', None),
                                       ('mean_val_score', None),
                                       ('sd_val_score', None),
                                       ('validation_meanerror', None),
                                       ('tr_label', None)])

    def test_allowed_par(self):
        par_dict = {'affinity': 'l1',
                    'linkage': 'ward'}
        par_dict_a = {'affinity': 'euclidean',
                      'linkage': 'ward'}
        self.assertFalse(self.paramselect._allowed_par(par_dict))
        self.assertTrue(self.paramselect._allowed_par(par_dict_a))


class TestPrivate(unittest.TestCase):

    def test_return_best(self):
        val_scores = [0.0, 0.5, 1.0]
        mult_val_scores = [0.0, 0.0, 0.5, 0.0]
        self.assertEqual(_return_best(val_scores), [0])
        self.assertEqual(_return_best(mult_val_scores), [0, 1, 3])

    def test_return_knownbest(self):
        val_scores = [0.0, 0.5, 1.0]
        idx = [1, 2]
        mult_val_scores = [0.0, 0.0, 0.5, 0.0]
        mult_idx = [0, 1]
        self.assertEqual(_return_knownbest(val_scores, idx), [1])
        self.assertEqual(_return_knownbest(mult_val_scores, mult_idx), [0, 1])

    def test_create_result_table(self):
        out = [(('s', KNeighborsClassifier()), ('s', LogisticRegression())),
               (('c', SpectralClustering(n_clusters=2)), ('c', SpectralClustering(n_clusters=2))),
               (('best_nclust', 2), ('best_nclust', 2)),
               (('mean_train_score', 0.0), ('mean_train_score', 0.0)),
               (('sd_train_score', 0.0), ('sd_train_score', 0.0)),
               (('mean_val_score', 0.0), ('mean_val_score', 0.0)),
               (('sd_val_score', 0.0), ('sd_val_score', 0.0)),
               (('validation_meanerror', (0.0, (0.0, 0.0))), ('validation_meanerror', (0.0, (0.0, 0.0))))]
        self.assertIsInstance(_create_result_table(out), dict)


if __name__ == 'main':
    unittest.main()



import time
from sklearn.datasets import make_blobs
from sklearn.model_selection import train_test_split
from sklearn.cluster import KMeans, AgglomerativeClustering, SpectralClustering
from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
from reval.best_nclust_cv import FindBestClustCV
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
import itertools
from hdbscan import HDBSCAN
import pickle as pkl
import matplotlib.pyplot as plt


def time_cmplx(n_jobs=1):
    data = make_blobs(100, 10, centers=2)
    data_tr, data_ts, y_tr, y_ts = train_test_split(data[0],
                                                    data[1],
                                                    test_size=0.5,
                                                    stratify=data[1],
                                                    random_state=42)
    s = [KNeighborsClassifier(), SVC(), LogisticRegression(), RandomForestClassifier()]
    c = [HDBSCAN(), AgglomerativeClustering(), KMeans(), SpectralClustering()]
    param = itertools.product(s, c)

    labels = ['KNN'] * 4 + ['SVM'] * 4 + ["LR"] * 4 + ['RF'] * 4
    time_cv = {'LR': [],
               'KNN': [],
               'RF': [],
               'SVM': []}
    time_ev = {'LR': [],
               'KNN': [],
               'RF': [],
               'SVM': []}
    for idx, mod in enumerate(param):
        classifier, clustering = mod[0], mod[1]
        findbest = FindBestClustCV(s=classifier,
                                   c=clustering,
                                   nrand=10,
                                   nfold=2,
                                   n_jobs=n_jobs,
                                   nclust_range=list(range(2, 7, 1)))
        if isinstance(clustering, HDBSCAN):
            start = time.time()
            _, _, tr_lab = findbest.best_nclust(data_tr, iter_cv=10, strat_vect=y_tr)
            time_cv[labels[idx]].append(time.time() - start)

            start = time.time()
            findbest.evaluate(data_tr, data_ts, nclust=2, tr_lab=tr_lab)
            time_ev[labels[idx]].append(time.time() - start)
        else:
            start = time.time()
            _, _ = findbest.best_nclust(data_tr, iter_cv=10, strat_vect=y_tr)
            time_cv[labels[idx]].append(time.time() - start)

            start = time.time()
            findbest.evaluate(data_tr, data_ts, nclust=2)
            time_ev[labels[idx]].append(time.time() - start)

    pkl.dump(time_cv, open(f'time_cv_njobs{n_jobs}.pkl', 'wb'))
    pkl.dump(time_ev, open(f'time_ev_njobs{n_jobs}.pkl', 'wb'))

    clustering = KMeans()
    classifier = KNeighborsClassifier()
    time_knnkmeans = {10: [],
                      100: [],
                      1000: []}
    for nsamples, nfeatures in itertools.product([100, 500, 1000, 1500, 2000, 2500, 3000],
                                                 [10, 100, 1000]):
        data = make_blobs(nsamples, nfeatures, centers=2)
        data_tr, data_ts, y_tr, y_ts = train_test_split(data[0],
                                                        data[1],
                                                        test_size=0.5,
                                                        stratify=data[1],
                                                        random_state=42)
        findbest = FindBestClustCV(s=classifier,
                                   c=clustering,
                                   nrand=10,
                                   nfold=2,
                                   n_jobs=n_jobs,
                                   nclust_range=list(range(2, 7, 1)))
        start = time.time()
        _, _ = findbest.best_nclust(data_tr, iter_cv=10, strat_vect=y_tr)
        findbest.evaluate(data_tr, data_ts, nclust=2)
        time_knnkmeans[nfeatures].append(time.time() - start)

    pkl.dump(time_knnkmeans, open(f'time_knnkmeans{n_jobs}.pkl', 'wb'))


def plot(file, save_name=None):
    time_dict = pkl.load(open(file, 'rb'))
    style = ['-', '-.', '--', ':']
    fig, ax = plt.subplots(figsize=(5, 5))
    cl_list = ['hdbscan', 'hierarchical', 'k-means', 'spectral']
    i = 0
    for s, t in time_dict.items():
        ax.plot(cl_list,
                t,
                linewidth=2,
                linestyle=style[i],
                label=s,
                color='black')
        i += 1
    ax.legend(fontsize=10, loc=2)
    plt.xticks(cl_list, fontsize=10)
    plt.yticks(fontsize=10)
    # plt.xlabel('Clustering algorithms', fontsize=10)
    plt.ylabel('Execution time (s)', fontsize=10, labelpad=15)
    if save_name is not None:
        plt.savefig(f'./{save_name}.pdf', format='pdf')
    else:
        plt.show()


def plot_one(file, save_name=None):
    time_one = pkl.load(open(file, 'rb'))
    style = ['-', '-.', '--']
    fig, ax = plt.subplots(figsize=(5, 5))
    cl_list = [100, 500, 1000, 1500, 2000, 2500, 3000]
    i = 0
    for s, t in time_one.items():
        ax.plot(cl_list,
                t,
                linewidth=2,
                linestyle=style[i],
                label=f'{s} features',
                color='black')
        i += 1
    ax.legend(fontsize=10, loc=2)
    plt.xticks(cl_list, fontsize=10)
    plt.yticks(fontsize=10)
    plt.xlabel('Number of samples', fontsize=10, labelpad=8)
    # plt.ylabel('Execution time (s)', fontsize=10)
    if save_name is not None:
        plt.savefig(f'./{save_name}.pdf', format='pdf')
    else:
        plt.show()


#
#
#
# setup = """
#
#
#
# s = KNeighborsClassifier()
# c = AgglomerativeClustering()
#
#
# # time_dict = {}
# # for classifier, clustering in param:
# findbest = FindBestClustCV(s=s,
#                            c=c,
#                            nrand=10,
#                            nfold=2,
#                            n_jobs=7,
#                            nclust_range=list(range(2, 7, 1)))
# """
#
# statement_1 = """
# # train_time, evalu_time = [], []
# # for it in range(100):
# # train_start = time.time()
# findbest.best_nclust(data_tr, iter_cv=10, strat_vect=y_tr)
# # train_time.append(time.time() - train_start)
# """
#
# statement_2 = """
# # evalu_start = time.time()
# findbest.evaluate(data_tr, data_ts, nclust=2)
# # evalu_time.append(time.time() - evalu_start)
# # time_dict[(classifier, clustering)] = [(np.mean(train_time), np.std(train_time)),
# #                                        (np.mean(evalu_time), np.std(evalu_time))]
#
# # for k, v in time_dict.items():
# #     print(f"S/C: {k} -- Execution time (tr/ts): {v}")
# """

if __name__ == "__main__":
    # time_cmplx(n_jobs=1)
    plot('time_cv_njobs1.pkl', save_name='time_cv_njobs1')
    plot('time_ev_njobs1.pkl', save_name='time_ev_njobs1')
    plot_one('time_knnkmeans1.pkl', save_name='time_knnkmeans_njobs1')

    # time_cmplx(n_jobs=7)
    plot('time_cv_njobs7.pkl', save_name='time_cv_njobs7')
    plot('time_ev_njobs7.pkl', save_name='time_ev_njobs7')
    plot_one('time_knnkmeans7.pkl', save_name='time_knnkmeans_njobs7')

# out_ev = timeit.repeat(stmt=statement_2, setup=setup, repeat=1)
# print(np.mean(out_cv), np.std(out_cv))
# print(np.mean(out_ev, np.std(out_ev)))

# time_cv = {'LR': ([2.59, 7.69, 8.60, 8.96], [0.10, 0.36, 0.66, 0.32]),
#            'KNN': ([2.05, 2.44, 3.22, 3.39], [0.19, 0.13, 0.20, 0.16]),
#            'RF': ([9.47, 30.63, 30.73, 41.49], [0.13, 1.43, 1.03, 6.09]),
#            'SVM': ([1.78, 2.00, 2.80, 2.92], [0.03, 0.06, 0.17, 0.07])}
# time_ev = {'LR': ([0.009, 0.006, 0.02, 0.05], [0.002, 0.0008, 0.003, 0.003]),
#            'KNN': ([0.008, 0.005, 0.02, 0.05], [0.0009, 0.0007, 0.002, 0.003]),
#            'RF': ([0.13, 0.1, 0.11, 0.19], [0.008, 0.006, 0.003, 0.02]),
#            'SVM': ([0.004, 0.002, 0.02, 0.04], [0.0005, 0.0003, 0.002, 0.002])}
# time_s = time_ev
# style = ['-', '-.', '--', ':']
# fig, ax = plt.subplots(figsize=(5, 5))
# cl_list = ['hdbscan', 'hierarchical', 'k-means', 'spectral']
# i = 0
# for s, time in time_s.items():
#     ax.plot(cl_list,
#             time[0],
#             linewidth=2,
#             linestyle=style[i],
#             label=s,
#             color='black')
#     i += 1
#     ax.errorbar(cl_list,
#                 time[0],
#                 [1.96 * (tval / 10) for tval in time[1]],
#                 linewidth=1,
#                 linestyle='-',
#                 label='',
#                 color='black')
# ax.legend(fontsize=10, loc=2)
# plt.xticks(cl_list, fontsize=10)
# plt.yticks(fontsize=10)
# plt.xlabel('Clustering algorithms', fontsize=10)
# plt.ylabel('Execution time (s)', fontsize=10)
# # if save_fig is not None:
# #     plt.savefig(f'./{save_fig}', format='png')
# # else:
# plt.show()



from sklearn.model_selection import RepeatedKFold, RepeatedStratifiedKFold
from reval.relative_validation import RelativeValidation
from collections import namedtuple
from scipy import stats
import numpy as np
import math
import logging
import multiprocessing as mp
import itertools
import pandas as pd

logging.basicConfig(format='%(asctime)s, %(levelname)s %(message)s',
                    datefmt='%H:%M:%S',
                    level=logging.INFO)


class FindBestClustCV(RelativeValidation):
    """
    Child class of :class:`reval.relative_validation.RelativeValidation`.
    It performs (repeated) k-fold cross validation on the training set to
    select the best number of clusters, i.e., the number that minimizes the
    normalized stability (i.e., average misclassification
    error/asymptotic misclassification rate).

    :param nfold: number of CV folds.
    :type nfold: int
    :param nclust_range: list with clusters to look for, default None.
    :type nclust_range: list of int
    :param s: classification object inherited
        from :class:`reval.relative_validation.RelativeValidation`.
    :type s: class
    :param c: clustering object inherited
        from :class:`reval.relative_validation.RelativeValidation`.
    :type c: class
    :param nrand: number of random labeling iterations
        to compute asymptotic misclassification rate, inherited from
        :class:`reval.relative_validation.RelativeValidation` class.
    :type nrand: int
    :param n_jobs: number of processes to be run in parallel, default 1.
    :type n_jobs: int

    :attribute: cv_results_ dataframe with cross validation results. Columns are
        ncl = number of clusters; ms_tr = misclassification training;
        ms_val = misclassification validation.
    """

    def __init__(self, s, c, nrand, nfold=2, n_jobs=1, nclust_range=None):
        """
        Construct method.
        """
        super().__init__(s, c, nrand)
        self.nfold = nfold
        if abs(n_jobs) > mp.cpu_count():
            self.n_jobs = mp.cpu_count()
        else:
            self.n_jobs = abs(n_jobs)
        self.nclust_range = nclust_range

    def best_nclust(self, data, iter_cv=1, strat_vect=None):
        """
        This method takes as input the training dataset and the
        stratification vector (if available) and performs a
        (repeated) CV procedure to select the best number of clusters that minimizes
        normalized stability.

        :param data: training dataset.
        :type data: ndarray, (n_samples, n_features)
        :param iter_cv: number of iteration for repeated CV, default 1.
        :type iter_cv: integer
        :param strat_vect: vector for stratification, defaults to None.
        :type strat_vect: ndarray, (n_samples,)
        :return: CV metrics for training and validation sets, best number of clusters,
            misclassification errors at each CV iteration.
        :rtype: dictionary, int, (list) if n_clusters parameter is not available
        """
        data_array = np.array(data)
        reval = RelativeValidation(self.class_method, self.clust_method, self.nrand)

        if strat_vect is not None:
            kfold = RepeatedStratifiedKFold(n_splits=self.nfold, n_repeats=iter_cv, random_state=42)
        else:
            kfold = RepeatedKFold(n_splits=self.nfold, n_repeats=iter_cv, random_state=42)
        fold_gen = kfold.split(data_array, strat_vect)
        if self.nclust_range is not None and 'n_clusters' in self.clust_method.get_params().keys():
            params = list(itertools.product([(data_array, reval)],
                                            fold_gen, self.nclust_range))
        else:
            params = list(itertools.product([(data_array, reval)], fold_gen))
        if self.n_jobs > 1:
            p = mp.Pool(processes=self.n_jobs)
            miscl = list(zip(*p.starmap(self._fit, params)))
            p.close()
            p.join()
            out = list(zip(*miscl))
        else:
            miscl = []
            for p in params:
                if len(p) > 2:
                    miscl.append(self._fit(data_obj=p[0],
                                           idxs=p[1],
                                           ncl=p[2]))
                else:
                    miscl.append(self._fit(data_obj=p[0],
                                           idxs=p[1]))
            out = miscl

        # return dataframe attribute (cv_results_) with cv scores
        # If no point are labeled (e.g., all points assigned to -1 class by HDBSCAN)
        # the method returns
        cv_results_ = pd.DataFrame(out,
                                   columns=['ncl', 'ms_tr', 'ms_val', 'tr_labels', 'val_labels'])
        ctrl_rows = cv_results_.shape[0]
        cv_results_.dropna(axis=0, inplace=True)
        if 0 < ctrl_rows - cv_results_.shape[0] < ctrl_rows:
            logging.info("Dropped results where clustering algorithm failed to identify clusters.")
            FindBestClustCV.cv_results_ = cv_results_
        elif ctrl_rows - cv_results_.shape[0] == 0:
            FindBestClustCV.cv_results_ = cv_results_
        else:
            logging.info(f"{self.clust_method} was not able to identify any cluster. Failed run with "
                         f"{self.class_method}.")
            return None

        metrics = {'train': {}, 'val': {}}
        for ncl in cv_results_.ncl.unique():
            norm_stab_tr = cv_results_.loc[cv_results_.ncl == ncl]['ms_tr']
            norm_stab_val = cv_results_.loc[cv_results_.ncl == ncl]['ms_val']
            metrics['train'][ncl] = (np.mean(norm_stab_tr), _confint(norm_stab_tr))
            metrics['val'][ncl] = (np.mean(norm_stab_val), _confint(norm_stab_val))

        val_score = np.array([val[0] for val in metrics['val'].values()])
        bestscore = min(val_score)
        # select the cluster with the minimum misclassification error
        # and the maximum number of clusters
        if self.nclust_range is not None and 'n_clusters' in self.clust_method.get_params().keys():
            bestncl = self.nclust_range[np.flatnonzero(val_score == bestscore)[-1]]
            return metrics, bestncl
        else:
            bestncl = list(metrics['val'].keys())[np.flatnonzero(val_score == bestscore)[-1]]
            best_idx = cv_results_.loc[cv_results_.ncl == bestncl].ms_val.idxmin()
            idx_vect = np.concatenate((params[best_idx][1][-2], params[best_idx][1][-1]))
            label_vect = np.concatenate((out[best_idx][-2], out[best_idx][-1]))
            tr_lab = [lab for _, lab in sorted(zip(idx_vect, label_vect))]
            return metrics, bestncl, tr_lab

    def evaluate(self, data_tr, data_ts, nclust=None, tr_lab=None):
        """
        Method that applies the selected clustering algorithm with the best number of clusters
        to the test set. It returns clustering labels.

        :param data_tr: training dataset.
        :type data_tr: ndarray, (n_samples, n_features)
        :param data_ts: test dataset.
        :type data_ts: ndarray, (n_samples, n_features)
        :param nclust: best number of clusters, default None.
        :type nclust: int
        :param tr_lab: clustering labels for the training set. If not None
            the clustering algorithm is not performed and the classifier is fitted.
            Available for clustering methods without `n_clusters` parameter. Default None.
        :type tr_lab: array-like
        :return: labels and accuracy for both training and test sets.
        :rtype: namedtuple, (train_cllab: array, train_acc:float, test_cllab:array, test_acc:float)
        """
        if isinstance(tr_lab, list):
            tr_lab = np.array(tr_lab)
        try:
            if 'n_clusters' in self.clust_method.get_params().keys():
                self.clust_method.n_clusters = nclust
                tr_misc, modelfit, labels_tr = super().train(data_tr)
            else:
                tr_misc, modelfit, labels_tr = super().train(data_tr, tr_lab)
            ts_misc, labels_ts = super().test(data_ts, modelfit)
        except TypeError:
            logging.info(f"Not possible to perform evaluation on the test set. "
                         f"No clusters were identified with "
                         f"{self.clust_method}")
            return None

        Eval = namedtuple('Eval',
                          ['train_cllab', 'train_acc', 'test_cllab', 'test_acc'])
        out = Eval(labels_tr, 1 - tr_misc, labels_ts, 1 - ts_misc)
        return out

    @staticmethod
    def _fit(data_obj, idxs, ncl=None):
        """
        Function that calls training, test, and random labeling.

        :param data_obj: dataset and `reval.RelativeValidation` class.
        :type data_obj: tuple
        :param idxs: lists of training and validation indices.
        :type idxs: tuple
        :param ncl: number of clusters, default None
        :type ncl: int
        :return: number of clusters and misclassification errors for training and validation.
        :rtype: tuple (int, float, float)
        """

        data_array, reval = data_obj
        tr_idx, val_idx = idxs
        tr_set, val_set = data_array[tr_idx], data_array[val_idx]
        if 'n_clusters' in reval.clust_method.get_params().keys():
            reval.clust_method.n_clusters = ncl

        # Case in which the number of clusters identified is not sufficient or points are not classified
        # (it happens particularly for SpectralClustering and HDBSCAN)
        try:
            miscl_tr, modelfit, tr_labels = reval.train(tr_set)
            miscl_val, val_labels = reval.test(val_set, modelfit)
            rndmisc_mean_val = reval.rndlabels_traineval(tr_set, val_set,
                                                         tr_labels,
                                                         val_labels)
            # If random labeling gives perfect prediction, substitute
            # misclassification for validation loop with 1.0
            if rndmisc_mean_val > 0.0:
                ms_val = miscl_val / rndmisc_mean_val
            else:
                ms_val = 1.0
            return len([n for n in np.unique(val_labels) if n >= 0]), miscl_tr, ms_val, tr_labels, val_labels
        except TypeError:
            return None, None, None, None, None


def _confint(vect):
    """
    Private function to compute confidence interval.

    :param vect: performance scores.
    :type vect: array-like
    :return: mean and error.
    :rtype: tuple
    """
    mean = np.mean(vect)
    # interval = 1.96 * math.sqrt((mean * (1 - mean)) / len(vect))
    interval = stats.t.ppf(1 - (0.05 / 2), len(vect) - 1) * (np.std(vect) / math.sqrt(len(vect)))
    return mean, interval



import cProfile
import pstats
from sklearn.datasets import make_blobs
from sklearn.model_selection import train_test_split
from sklearn.cluster import KMeans, AgglomerativeClustering
from hdbscan import HDBSCAN
from sklearn.neighbors import KNeighborsClassifier
from reval.best_nclust_cv import FindBestClustCV

s = KNeighborsClassifier()
# c = KMeans()
c = AgglomerativeClustering()

findbest = FindBestClustCV(s=s,
                           c=c,
                           nrand=10,
                           nfold=2,
                           n_jobs=1,
                           nclust_range=list(range(2, 7, 1)))

data = make_blobs(100, 10, centers=2)
data_tr, data_ts, y_tr, y_ts = train_test_split(data[0],
                                                data[1],
                                                test_size=0.5,
                                                stratify=data[1],
                                                random_state=42)

print("Profiling of algorithm that finds the best number of clusters.")
profiler = cProfile.Profile()
profiler.enable()
findbest.best_nclust(data_tr, iter_cv=10, strat_vect=y_tr)
profiler.disable()
stats = pstats.Stats(profiler).sort_stats('ncalls')
stats.print_stats()

print("Profiling of algorithm that evaluates the best number of clusters.")
profiler.enable()
findbest.evaluate(data_tr, data_ts, nclust=2)
profiler.disable()
stats = pstats.Stats(profiler).sort_stats('ncalls')
stats.print_stats()



import sys

sys.path.append('../../reval_clustering/')

from reval.best_nclust_cv import FindBestClustCV
from reval.internal_baselines import select_best, evaluate_best
from reval.visualization import plot_metrics
from reval.utils import kuhn_munkres_algorithm, compute_metrics
from reval.param_selection import SCParamSelection
from datasets.manuscript_builddatasets import build_ucidatasets

from sklearn.datasets import make_blobs, load_digits
from sklearn.model_selection import train_test_split
from sklearn.neighbors import KNeighborsClassifier
from sklearn.cluster import AgglomerativeClustering, KMeans
from sklearn.metrics import adjusted_mutual_info_score, silhouette_score, davies_bouldin_score
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
from sklearn.datasets import fetch_openml
from sklearn.preprocessing import StandardScaler
import hdbscan
from umap import UMAP
import numpy as np

import warnings
import logging
import matplotlib.pyplot as plt

# Modify this variable for parallelization
N_JOBS = 7

warnings.filterwarnings('ignore')

"""
Three example functions that can also be run from shell (see manuscript_examples.py file). 

Example 1: blobs dataset;
Example 2: hand-written digits dataset from UCI repository;
Example 3: real-world dataset MNIST;
Example 4: best clussifier-clustering combinations for 18 datasets from UCI Machine Learning Repository.
"""


# EXAMPLE 1: Isotropic Gaussian blobs
def example1():
    # Generate dataset
    data = make_blobs(1000, 2, centers=5,
                      center_box=(-20, 20),
                      random_state=42)

    # Visualize dataset
    plt.figure(figsize=(6, 4))
    for i in range(5):
        plt.scatter(data[0][data[1] == i][:, 0],
                    data[0][data[1] == i][:, 1],
                    label=i, cmap='tab20')
    plt.title("Blobs dataset")
    # plt.savefig('./blobs.png', format='png')
    plt.show()

    # Create training and test sets
    X_tr, X_ts, y_tr, y_ts = train_test_split(data[0],
                                              data[1],
                                              test_size=0.30,
                                              random_state=42,
                                              stratify=data[1])

    # Initialize clustering and classifier
    classifier = KNeighborsClassifier(n_neighbors=15)
    clustering = KMeans()

    # Run relatve validation (repeated CV and testing)
    findbestclust = FindBestClustCV(nfold=2,
                                    nclust_range=list(range(2, 7, 1)),
                                    s=classifier,
                                    c=clustering,
                                    nrand=10,
                                    n_jobs=N_JOBS)
    metrics, nbest = findbestclust.best_nclust(X_tr, iter_cv=10, strat_vect=y_tr)
    out = findbestclust.evaluate(X_tr, X_ts, nclust=nbest)

    # Plot CV metrics
    plot_metrics(metrics, prob_lines=False)
    logging.info(f"Validation stability: {metrics['val'][nbest]}")
    perm_lab = kuhn_munkres_algorithm(y_ts, out.test_cllab)

    logging.info(f"Best number of clusters: {nbest}")
    logging.info(f'AMI (true labels vs predicted labels) for test set = '
                 f'{adjusted_mutual_info_score(y_ts, out.test_cllab)}')
    logging.info('\n\n')

    # Compute metrics
    logging.info("Metrics from true label comparisons on test set:")
    class_scores = compute_metrics(y_ts, perm_lab, perm=False)
    for k, val in class_scores.items():
        if k in ['F1', 'MCC']:
            logging.info(f"{k}, {val}")
    logging.info("\n\n")

    # Internal measures
    # SILHOUETTE
    logging.info("Silhouette score based selection")
    sil_score_tr, sil_best_tr, sil_label_tr = select_best(X_tr, clustering, silhouette_score,
                                                          select='max',
                                                          nclust_range=list(range(2, 7, 1)))
    sil_score_ts, sil_best_ts, sil_label_ts = select_best(X_ts, clustering, silhouette_score,
                                                          select='max',
                                                          nclust_range=list(range(2, 7, 1)))

    sil_eval = evaluate_best(X_ts, clustering, silhouette_score, sil_best_tr)

    logging.info(f"Best number of clusters (and scores) for tr/ts independent runs: "
                 f"{sil_best_tr}({sil_score_tr})/{sil_best_ts}({sil_score_ts})")
    logging.info(f"Test set evaluation {sil_eval}")
    logging.info(f'AMI (true labels vs clustering labels) training = '
                 f'{adjusted_mutual_info_score(y_tr, kuhn_munkres_algorithm(y_tr, sil_label_tr))}')
    logging.info(f'AMI (true labels vs clustering labels) test = '
                 f'{adjusted_mutual_info_score(y_ts, kuhn_munkres_algorithm(y_ts, sil_label_ts))}')
    logging.info('\n\n')

    # DAVIES-BOULDIN
    logging.info("Davies-Bouldin score based selection")
    db_score_tr, db_best_tr, db_label_tr = select_best(X_tr, clustering, davies_bouldin_score,
                                                       select='min', nclust_range=list(range(2, 7, 1)))
    db_score_ts, db_best_ts, db_label_ts = select_best(X_ts, clustering, davies_bouldin_score,
                                                       select='min', nclust_range=list(range(2, 7, 1)))

    db_eval = evaluate_best(X_ts, clustering, davies_bouldin_score, db_best_tr)

    logging.info(f"Best number of clusters (and scores) for tr/ts independent runs: "
                 f"{db_best_tr}({db_score_tr})/{db_best_ts}({db_score_ts})")
    logging.info(f"Test set evaluation {db_eval}")
    logging.info(f'AMI (true labels vs clustering labels) training = '
                 f'{adjusted_mutual_info_score(y_tr, kuhn_munkres_algorithm(y_tr, db_label_tr))}')
    logging.info(f'AMI (true labels vs clustering labels) test = '
                 f'{adjusted_mutual_info_score(y_ts, kuhn_munkres_algorithm(y_ts, db_label_ts))}')
    logging.info('\n\n')

    # Plot true vs predicted labels for test sets
    plt.figure(figsize=(6, 4))
    for i in range(5):
        plt.scatter(X_ts[y_ts == i][:, 0],
                    X_ts[y_ts == i][:, 1],
                    label=str(i),
                    cmap='tab20')
    plt.legend(loc=3)
    plt.title("Test set true labels")
    # plt.savefig('./blobs_true.png', format='png')
    plt.show()

    plt.figure(figsize=(6, 4))
    for i in range(5):
        plt.scatter(X_ts[perm_lab == i][:, 0],
                    X_ts[perm_lab == i][:, 1],
                    label=str(i),
                    cmap='tab20')
    plt.legend(loc=3)
    plt.title("Test set clustering labels")
    # plt.savefig('./blobs_clustering.png', format='png')
    plt.show()


# Example 2: MNIST dataset
def example2():
    mnist = fetch_openml('mnist_784', version=1)
    mnist.target = mnist.target.astype(int)

    X_tr, y_tr = mnist['data'][:60000], mnist.target[:60000]
    X_ts, y_ts = mnist['data'][60000::], mnist.target[60000::]
    transform = UMAP(n_components=2,
                     random_state=42,
                     n_neighbors=30,
                     min_dist=0.0)
    X_tr = transform.fit_transform(X_tr)
    X_ts = transform.transform(X_ts)

    s = KNeighborsClassifier(n_neighbors=30)
    c = hdbscan.HDBSCAN(min_samples=10,
                        min_cluster_size=200)

    reval = FindBestClustCV(s=s,
                            c=c,
                            nfold=2,
                            nrand=10,
                            n_jobs=N_JOBS)

    metrics, nclustbest, tr_lab = reval.best_nclust(X_tr, iter_cv=10, strat_vect=y_tr)

    plot_metrics(metrics)

    out = reval.evaluate(X_tr, X_ts, nclust=nclustbest, tr_lab=tr_lab)
    perm_lab = kuhn_munkres_algorithm(y_ts, out.test_cllab)
    logging.info(f"Validation stability: {metrics['val'][nclustbest]}")

    logging.info(f"Best number of clusters during CV: {nclustbest}")
    logging.info(f"Best number of clusters on test set: "
                 f"{len([lab for lab in np.unique(out.test_cllab) if lab >= 0])}")
    logging.info(f'AMI (true labels vs predicted labels) = '
                 f'{adjusted_mutual_info_score(y_ts, out.test_cllab)}')
    logging.info('\n\n')

    logging.info("Metrics from true label comparisons on test set:")
    class_scores = compute_metrics(y_ts, perm_lab)
    for k, val in class_scores.items():
        logging.info(f'{k}, {val}')
    logging.info('\n\n')

    # Visualization
    fig, ax = plt.subplots(figsize=(10, 8))
    scatter = ax.scatter(X_tr[:, 0],
                         X_tr[:, 1],
                         c=y_tr, cmap='rainbow_r',
                         s=0.1)
    legend = ax.legend(*scatter.legend_elements())
    ax.add_artist(legend)
    plt.title("Train set true labels (digits dataset)")
    plt.show()

    fig, ax = plt.subplots(figsize=(10, 8))
    scatter = ax.scatter(X_tr[:, 0],
                         X_tr[:, 1],
                         c=kuhn_munkres_algorithm(y_tr, tr_lab),
                         cmap='tab20',
                         s=0.1)
    legend = ax.legend(*scatter.legend_elements())
    ax.add_artist(legend)
    plt.title("Train set predicted labels (digits dataset)")
    plt.show()

    fig, ax = plt.subplots(figsize=(10, 8))
    scatter = ax.scatter(X_ts[:, 0],
                         X_ts[:, 1],
                         c=y_ts, cmap='tab20',
                         s=0.1)
    legend = ax.legend(*scatter.legend_elements())
    ax.add_artist(legend)
    plt.title("Test set true labels (digits dataset)")
    plt.show()

    fig, ax = plt.subplots(figsize=(10, 8))
    scatter = ax.scatter(X_ts[:, 0],
                         X_ts[:, 1],
                         s=0.1,
                         c=perm_lab, cmap='tab20')
    legend = ax.legend(*scatter.legend_elements())
    ax.add_artist(legend)
    plt.title("Test set clustering labels (digits dataset)")
    plt.show()

    # Internal measures
    # SILHOUETTE
    logging.info("Silhouette score based selection")
    sil_score_tr, sil_best_tr, sil_label_tr = select_best(X_tr, c, silhouette_score, select='max')
    sil_score_ts, sil_best_ts, sil_label_ts = select_best(X_ts, c, silhouette_score, select='max')
    logging.info(
        f"Best number of clusters (and scores) for tr/ts independent runs: "
        f"{sil_best_tr}({sil_score_tr})/{sil_best_ts}({sil_score_ts})")
    logging.info(f'AMI (true labels vs clustering labels) training = '
                 f'{adjusted_mutual_info_score(y_tr, kuhn_munkres_algorithm(y_tr, sil_label_tr))}')
    logging.info(f'AMI (true labels vs clustering labels) test = '
                 f'{adjusted_mutual_info_score(y_ts, kuhn_munkres_algorithm(y_ts, sil_label_ts))}')
    logging.info('\n\n')

    # DAVIES-BOULDIN
    logging.info("Davies-Bouldin score based selection")
    db_score_tr, db_best_tr, db_label_tr = select_best(X_tr, c, davies_bouldin_score,
                                                       select='min')
    db_score_ts, db_best_ts, db_label_ts = select_best(X_ts, c, davies_bouldin_score,
                                                       select='min')

    logging.info(
        f"Best number of clusters (and scores) for tr/ts independent runs: "
        f"{db_best_tr}({db_score_tr})/{db_best_ts}({db_score_ts})")
    logging.info(f'AMI (true labels vs clustering labels) training = '
                 f'{adjusted_mutual_info_score(y_tr, kuhn_munkres_algorithm(y_tr, db_label_tr))}')
    logging.info(f'AMI (true labels vs clustering labels) test = '
                 f'{adjusted_mutual_info_score(y_ts, kuhn_munkres_algorithm(y_ts, db_label_ts))}')
    logging.info('\n\n')

    # Visualization
    fig, ax = plt.subplots(figsize=(10, 8))
    scatter = ax.scatter(X_tr[:, 0],
                         X_tr[:, 1],
                         c=sil_label_tr, cmap='tab20',
                         s=0.1)
    legend = ax.legend(*scatter.legend_elements())
    ax.add_artist(legend)
    plt.title("Train set silhouette labels (digits dataset)")
    plt.show()

    fig, ax = plt.subplots(figsize=(10, 8))
    scatter = ax.scatter(X_ts[:, 0],
                         X_ts[:, 1],
                         c=sil_label_ts, cmap='tab20',
                         s=0.1)
    legend = ax.legend(*scatter.legend_elements())
    ax.add_artist(legend)
    legend = ax.legend(*scatter.legend_elements())
    ax.add_artist(legend)
    plt.title("Test set silhouette labels (digits dataset)")
    plt.show()

    fig, ax = plt.subplots(figsize=(10, 8))
    scatter = ax.scatter(X_tr[:, 0],
                         X_tr[:, 1],
                         c=db_label_tr, cmap='tab20',
                         s=0.1)
    legend = ax.legend(*scatter.legend_elements())
    ax.add_artist(legend)
    plt.title("Train set Davies-Bouldin labels (digits dataset)")
    plt.show()

    fig, ax = plt.subplots(figsize=(10, 8))
    scatter = ax.scatter(X_ts[:, 0],
                         X_ts[:, 1],
                         s=0.1,
                         c=db_label_ts, cmap='tab20')
    legend = ax.legend(*scatter.legend_elements())
    ax.add_artist(legend)
    plt.title("Test set Davies-Bouldin labels (digits dataset)")
    plt.show()


# Example 3: best classifier/clustering combination for UCI dataset
def example3(n_jobs, preprocess=None):
    """
    :param preprocess: it can be 'scaled',
        'umap', 'scaled+umap', default None for raw processing.
    :type preprocess: str
    :return:
    """
    # Example 4: best clussifier/clustering for UCI dataset

    # Classifiers
    s = [LogisticRegression(solver='liblinear',
                            random_state=42),
         RandomForestClassifier(n_estimators=100,
                                random_state=42),
         KNeighborsClassifier(n_neighbors=1,
                              metric='euclidean'),
         SVC(C=1,
             random_state=42)]

    # Clustering
    c = [AgglomerativeClustering(),
         KMeans(random_state=42),
         hdbscan.HDBSCAN()]

    scparam = {'s': s,
               'c': c}

    transform = UMAP(n_neighbors=30, min_dist=0.0, random_state=42)
    scale = StandardScaler()

    # Import benchmark datasets
    uci_data = build_ucidatasets()
    # Run ensemble learning algorithm
    best_results = {}
    for data, name in zip(uci_data, uci_data._fields):
        scparam['s'][-1].gamma = (1 / data['data'].shape[0])
        nclass = len(np.unique(data['target']))
        logging.info(f"Processing dataset {name}")
        logging.info(f"True number of classes: {nclass}\n")
        X_tr, X_ts, y_tr, y_ts = train_test_split(data['data'],
                                                  data['target'],
                                                  test_size=0.40,
                                                  random_state=42,
                                                  stratify=data['target'])
        if preprocess == 'umap+scaled':
            X_tr = transform.fit_transform(scale.fit_transform(X_tr))
        elif preprocess == 'umap':
            X_tr = transform.fit_transform(X_tr)
        elif preprocess == 'scaled':
            X_tr = scale.fit_transform(X_tr)

        scparam_select = SCParamSelection(sc_params=scparam,
                                          cv=2,
                                          nrand=10,
                                          clust_range=list(range(2, nclass + 3, 1)),
                                          n_jobs=n_jobs,
                                          iter_cv=10,
                                          strat=y_tr)
        scparam_select.fit(X_tr, nclass=nclass)
        best_results[name] = scparam_select.best_param_
        # Uncomment to save the results
        #     pkl.dump(best_results, open('./best_resultUCI_scaledumap.pkl', 'wb'))
        logging.info('*' * 100)
        logging.info('\n\n')


if __name__ == "__main__":
    # Format logging output and save
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    logging.basicConfig(format='%(asctime)s, %(levelname)s %(message)s',
                        datefmt='%H:%M:%S',
                        level=logging.INFO,
                        filename='example_out.log',
                        filemode='w')
    example1()
    example2()
    example3(n_jobs=7, preprocess='scaled+umap')



import numpy as np


def select_best(data, c, int_measure, select='max', nclust_range=None):
    """
    Select the best number of clusters that minimizes/maximizes
    the internal measure selected.

    :param data: dataset.
    :type data: array-like
    :param c: clustering algorithm class.
    :type c: obj
    :param int_measure: internal measure function.
    :type int_measure: obj
    :param select: it can be 'min', if the internal measure is to be minimized
        or 'max' if the internal measure should be macimized.
    :type select: str
    :param nclust_range: Range of clusters to consider, default None.
    :type nclust_range: list
    :return: internal score and best number of clusters.
    :rtype: float, int
    """
    if nclust_range is not None:
        scores = []
        label_vect = []
        for nc in nclust_range:
            c.n_clusters = nc
            label = c.fit_predict(data)
            scores.append(int_measure(data, label))
            label_vect.append(label)
    else:
        label = c.fit_predict(data)
        best = int_measure(data, label)
        return best, len([lab for lab in np.unique(label) if lab >= 0]), label
    if select == 'max':
        best = np.where(np.array(scores) == max(scores))[0]
    elif select == 'min':
        best = np.where(np.array(scores) == min(scores))[0]
    if len(set(label_vect[int(max(best))])) == nclust_range[int(max(best))]:
        return scores[int(max(best))], nclust_range[int(max(best))], label_vect[int(max(best))]
    else:
        return scores[int(max(best))], len(set(label_vect[int(max(best))])), label_vect[int(max(best))]


def evaluate_best(data, c, int_measure, ncl=None):
    """
    Function that, given a number of clusters, returns the corresponding internal measure
    for a dataset.

    :param data: dataset.
    :type data: array-like
    :param c: clustering algorithm class.
    :type c: obj
    :param int_measure: internal measure function.
    :type int_measure: obj
    :param ncl: number of clusters.
    :type ncl: int
    :return: internal score.
    :rtype: float
    """
    if 'n_clusters' in c.get_params().keys():
        c.n_clusters = ncl
    label = c.fit_predict(data)
    return int_measure(data, label)



from reval.best_nclust_cv import FindBestClustCV
from reval.relative_validation import RelativeValidation
from sklearn.model_selection import ParameterGrid
import multiprocessing as mp
import logging
import numpy as np
import itertools

from sklearn.cluster import AgglomerativeClustering

logging.basicConfig(format='%(asctime)s, %(levelname)s %(message)s',
                    datefmt='%H:%M:%S',
                    level=logging.INFO)


class SCParamSelection:
    """
    Class that implements grid search cross-validation in parallel to select the best
    combination of classifier/clustering methods.

    :param sc_params: dictionary of the form {'s': list, 'c': list} including the lists
        of classifiers and clustering methods to fit to the data.
    :type sc_params: dict
    :param cv: cross-validation folds.
    :type cv: int
    :param nrand: number of random label iterations.
    :type nrand: int
    :param n_jobs: number of jobs to run in parallel, default (number of cpus - 1).
    :type n_jobs: int
    :param iter_cv: number of repeated cv, default 1.
    :type iter_cv: int
    :param clust_range: list with number of clusters to investigate, default None.
    :type clust_range: list
    :param strat: stratification vector for cross-validation splits, default None.
    :type strat: numpy array

    :attribute: cv_results_ cross-validation results that can be directly transformed to
        a dataframe. Key names: 's', 'c', 'best_nclust', 'mean_train_score', 'sd_train_score',
        'mean_val_score', 'sd_val_score', 'validation_meanerror'. Dictionary of lists.
    :attribute: best_param_ best solution(s) selected (minimum validation error). List.
    :attribute: best_index_ index/indices of the best solution(s). Values correspond to the
        rows of the `cv_results_` table. List.
    """

    def __init__(self, sc_params, cv, nrand,
                 n_jobs,
                 iter_cv=1,
                 clust_range=None,
                 strat=None):
        self.sc_params = sc_params
        if len(self.sc_params['s']) == 1 and len(self.sc_params['c']) == 1:
            raise AttributeError("Please add at least another classifier/clustering "
                                 "method to run the parameter selection.")
        self.cv = cv
        self.nrand = nrand
        self.clust_range = clust_range
        if abs(n_jobs) > mp.cpu_count():
            self.n_jobs = mp.cpu_count()
        else:
            self.n_jobs = abs(n_jobs)
        self.iter_cv = iter_cv
        self.strat = strat

    def fit(self, data_tr, nclass=None):
        """
        Class method that performs grid search cross-validation on training data. If the number of
        true classes is known, the method returns both the best result with the correct number of
        clusters (and minimum stability), if available, and the overall best result (overall minimum stability).
        The output reports None if the clustering algorithm does not find any cluster (e.g., HDBSCAN label
        all points as -1).

        :param data_tr: training dataset.
        :type data_tr: numpy array
        :param nclass: number of true classes, default None.
        :type nclass: int
        """
        sc_grid = list(ParameterGrid(self.sc_params))
        params = list(zip([data_tr] * len(sc_grid), sc_grid))

        logging.info(f'Running {len(params)} combinations of '
                     f'classification/clustering methods...\n')

        p = mp.Pool(processes=self.n_jobs)
        out = list(zip(*p.starmap(self._run_gridsearchcv, params)))
        p.close()
        p.join()

        # cv_results_
        res_dict = _create_result_table(out)
        SCParamSelection.cv_results_ = res_dict

        # best_param_, best_index_
        val_scores = [vs for vs in res_dict['mean_val_score'] if vs is not None]
        val_idx = [idx for idx, vs in enumerate(res_dict['mean_val_score']) if vs is not None]
        if len(val_scores) > 0:
            idx_best = [val_idx[i] for i in _return_best(val_scores)]
        else:
            logging.info(f"No clustering solutions were found with any parameter combinations.")
            return self

        out_best = []
        if nclass is not None:
            logging.info(f'True number of clusters known: {nclass}\n')
            idx = np.where(np.array(res_dict['best_nclust']) == nclass)[0]
            idx_inter = set(idx).intersection(set(idx_best))
            if len(idx_inter) > 0:
                idx_best = list(idx_inter)
            else:
                if len(idx) > 0:
                    idx_true = _return_knownbest(res_dict['mean_val_score'], idx)
                    logging.info(f'Best solution(s) with true number of clusters:')
                    for bidx in idx_true:
                        logging.info(f'Models S/C: {res_dict["s"][bidx]}/{res_dict["c"][bidx]})')
                        logging.info(f'Validation performance: {res_dict["validation_meanerror"][bidx]}')
                        logging.info(f'N clusters: {res_dict["best_nclust"][bidx]}\n')
                        out_best.append([res_dict["s"][bidx], res_dict["c"][bidx],
                                         res_dict["best_nclust"][bidx], res_dict["validation_meanerror"][bidx]])
        logging.info(f'Best solution(s):')
        for bidx in idx_best:
            logging.info(f'Models: {res_dict["s"][bidx]}/{res_dict["c"][bidx]})')
            logging.info(f'Validation performance: {res_dict["validation_meanerror"][bidx]}')
            logging.info(f'N clusters: {res_dict["best_nclust"][bidx]}\n')
            out_best.append([res_dict["s"][bidx], res_dict["c"][bidx],
                             res_dict["best_nclust"][bidx], res_dict["validation_meanerror"][bidx]])
        SCParamSelection.best_param_ = out_best
        SCParamSelection.best_index_ = idx_best

        return self

    def _run_gridsearchcv(self, data, sc):
        """
        Private function with different initializations of
        :class:`reval.best_nclust_cv.FindBestClustCV`.

        :param data: input dataset.
        :type data: numpy array
        :param sc: classifier/clustering of the form {'s':, 'c':}.
        :type sc: dict
        :return: performance list.
        :rtype: list
        """
        findclust = FindBestClustCV(s=sc['s'],
                                    c=sc['c'],
                                    nfold=self.cv,
                                    nrand=self.nrand,
                                    n_jobs=1,
                                    nclust_range=self.clust_range)

        if 'n_clusters' in sc['c'].get_params().keys():
            metric, nclbest = findclust.best_nclust(data, iter_cv=self.iter_cv, strat_vect=self.strat)
            sc['c'].n_clusters = nclbest
            tr_lab = None
        else:
            try:
                metric, nclbest, tr_lab = findclust.best_nclust(data, iter_cv=self.iter_cv, strat_vect=self.strat)
            except TypeError:
                perf = [('s', sc['s']), ('c', sc['c']), ('best_nclust', None),
                        ('mean_train_score', None),
                        ('sd_train_score', None),
                        ('mean_val_score', None),
                        ('sd_val_score', None),
                        ('validation_meanerror', None),
                        ('tr_label', None)]
                return perf

        cv_scores = findclust.cv_results_
        perf = [('s', sc['s']), ('c', sc['c']), ('best_nclust', nclbest),
                ('mean_train_score', np.mean(cv_scores.loc[cv_scores.ncl == nclbest]['ms_tr'])),
                ('sd_train_score', np.std(cv_scores.loc[cv_scores.ncl == nclbest]['ms_tr'])),
                ('mean_val_score', np.mean(cv_scores.loc[cv_scores.ncl == nclbest]['ms_val'])),
                ('sd_val_score', np.std(cv_scores.loc[cv_scores.ncl == nclbest]['ms_val'])),
                ('validation_meanerror', metric['val'][nclbest]),
                ('tr_label', tr_lab)]
        return perf


class ParamSelection(RelativeValidation):
    """
    Class that implements grid search cross-validation in parallel to select
    the best combinations of parameters for fixed classifier/clustering algorithms.

    :param params: dictionary of dictionaries of the form {'s': {classifier parameter grid},
        'c': {clustering parameter grid}}. If one of the two dictionary of parameters is not
        available, initialize key but leave dictionary empty.
    :type params: dict
    :param cv: cross-validation folds.
    :type cv: int
    :param clust_range: list with number of clusters to investigate.
    :type clust_range: list
    :param n_jobs: number of jobs to run in parallel, default (number of cpus - 1).
    :type n_jobs: int
    :param iter_cv: number of repeated cv loops, default 1.
    :type iter_cv: int
    :param strat: stratification vector for cross-validation splits, default None.
    :type strat: numpy array

    :attribute: cv_results_ cross-validation results that can be directly transformed to
        a dataframe. Key names: classifier parameters, clustering parameters,
        'best_nclust', 'mean_train_score', 'sd_train_score',
        'mean_val_score', 'sd_val_score', 'validation_meanerror'. Dictionary of lists.
    :attribute: best_param_ best solution(s) selected (minimum validation error). List.
    :attribute: best_index_ index/indices of the best solution(s). Values correspond to the
        rows of the `cv_results_` table. List.
    """

    def __init__(self, params, cv, s, c, nrand,
                 n_jobs, iter_cv=1, strat=None, clust_range=None):
        super().__init__(s, c, nrand)
        self.params = params
        self.cv = cv
        self.iter_cv = iter_cv
        self.clust_range = clust_range
        if abs(n_jobs) > mp.cpu_count():
            self.n_jobs = mp.cpu_count()
        else:
            self.n_jobs = abs(n_jobs)
        self.strat = strat

    def fit(self, data_tr, nclass=None):
        """
        Class method that performs grid search cross-validation on training data. It
        deals with the error due to wrong parameter combinations (e.g., ward linkage
        with no euclidean affinity). If the true number of classes is know, the method
        selects both the best parameter combination that selects the true number of clusters
        (minimum stability) and the best parameter combination that minimizes
        overall stability.

        :param data_tr: training dataset.
        :type data_tr: numpy array
        :param nclass: number of true classes, default None.
        :type nclass: int
        """
        grid = {'s': ParameterGrid(self.params['s']), 'c': ParameterGrid(self.params['c'])}
        new_grid = list(itertools.product(grid['s'], grid['c']))
        new_params = [(data_tr, ng[0], ng[1]) for ng in new_grid if self._allowed_par(ng[1])]

        if len(new_grid) != len(new_params):
            logging.info(f"Dropped {len(new_grid) - len(new_params)} out of {len(new_grid)} parameter "
                         f"combinations "
                         f"due to {self.clust_method} class requirements.")

        logging.info(f'Running {len(new_params)} combinations of '
                     f'parameters...\n')

        p = mp.Pool(processes=self.n_jobs)
        out = list(zip(*p.starmap(self._run_gridsearchcv, new_params)))
        p.close()
        p.join()

        # cv_results_
        res_dict = _create_result_table(out)
        ParamSelection.cv_results_ = res_dict

        # best_param_, best_index_
        val_scores = [vs for vs in res_dict['mean_val_score'] if vs is not None]
        val_idx = [idx for idx, vs in enumerate(res_dict['mean_val_score']) if vs is not None]
        if len(val_scores) > 0:
            idx_best = [val_idx[i] for i in _return_best(val_scores)]
        else:
            logging.info(f"No clustering solutions were found with any parameter combinations.")
            return self

        out_best = []
        if nclass is not None:
            logging.info(f'True number of clusters known: {nclass}\n')
            idx = np.where(np.array(res_dict['best_nclust']) == nclass)[0]
            idx_inter = set(idx).intersection(set(idx_best))
            if len(idx_inter) > 0:
                idx_best = list(idx_inter)
            else:
                if len(idx) > 0:
                    idx_true = _return_knownbest(res_dict['mean_val_score'], idx)
                    logging.info(f'Best solution(s) with true number of clusters:')
                    for bidx in idx_true:
                        for k in self.params['s'].keys():
                            logging.info(f'Parameters classifier (S): {k}={res_dict[k][bidx]}')
                        for k in self.params['c'].keys():
                            logging.info(f'Parameters clustering (C): {k}={res_dict[k][bidx]}')
                        logging.info(f'Validation performance: {res_dict["validation_meanerror"][bidx]}')
                        logging.info(f'N clusters: {res_dict["best_nclust"][bidx]}\n')
                        out_best.append([res_dict[k][bidx] for k in self.params['s'].keys()] +
                                        [res_dict[k][bidx] for k in self.params['s'].keys()] +
                                        [res_dict["best_nclust"][bidx], res_dict["validation_meanerror"][bidx]])
        logging.info(f'Best solution(s):')
        for bidx in idx_best:
            for k in self.params['s'].keys():
                logging.info(f'Parameters classifier (S): {k}={res_dict[k][bidx]}')
            for k in self.params['c'].keys():
                logging.info(f'Parameters clustering (C): {k}={res_dict[k][bidx]}')
            logging.info(f'Validation performance: {res_dict["validation_meanerror"][bidx]}')
            logging.info(f'N clusters: {res_dict["best_nclust"][bidx]}\n')
            out_best.append([res_dict[k][bidx] for k in self.params['s'].keys()] +
                            [res_dict[k][bidx] for k in self.params['s'].keys()] +
                            [res_dict["best_nclust"][bidx], res_dict["validation_meanerror"][bidx]])
        ParamSelection.best_param_ = out_best
        ParamSelection.best_index_ = idx_best

        return self

    def _run_gridsearchcv(self, data, param_s, param_c):
        """
        Private method that initializes classifier/clustering with different
        parameter combinations and :class:`reval.best_nclust_cv.FindBestClustCV`.

        :param data: training dataset.
        :type data: numpy array
        :param param_s: dictionary of classifier parameters.
        :type: dict
        :param param_c: dictionary of clustering parameters.
        :type param_c: dict
        :return: performance list.
        :rtype: list
        """
        self.class_method.set_params(**param_s)
        self.clust_method.set_params(**param_c)
        findclust = FindBestClustCV(nfold=self.cv,
                                    s=self.class_method,
                                    c=self.clust_method,
                                    nrand=self.nrand, n_jobs=1,
                                    nclust_range=self.clust_range)
        if self.clust_range is not None:
            metric, nclbest = findclust.best_nclust(data, iter_cv=self.iter_cv, strat_vect=self.strat)
            tr_lab = None
        else:
            try:
                metric, nclbest, tr_lab = findclust.best_nclust(data, iter_cv=self.iter_cv, strat_vect=self.strat)
            except TypeError:
                perf = [(key, val) for key, val in param_s.items()] + \
                       [(key, val) for key, val in param_c.items()] + \
                       [('best_nclust', None),
                        ('mean_train_score', None),
                        ('sd_train_score', None),
                        ('mean_val_score', None),
                        ('sd_val_score', None),
                        ('validation_meanerror', None),
                        ('tr_label', None)]
                return perf

        perf = [(key, val) for key, val in param_s.items()] + \
               [(key, val) for key, val in param_c.items()] + \
               [('best_nclust', nclbest),
                ('mean_train_score', np.mean(
                    findclust.cv_results_.loc[findclust.cv_results_.ncl == nclbest]['ms_tr'])),
                ('sd_train_score', np.std(
                    findclust.cv_results_.loc[findclust.cv_results_.ncl == nclbest]['ms_tr'])),
                ('mean_val_score', np.mean(
                    findclust.cv_results_.loc[findclust.cv_results_.ncl == nclbest]['ms_val'])),
                ('sd_val_score', np.std(
                    findclust.cv_results_.loc[findclust.cv_results_.ncl == nclbest]['ms_val'])),
                ('validation_meanerror', metric['val'][nclbest]),
                ('tr_label', tr_lab)]
        return perf

    def _allowed_par(self, par_dict):
        """
        Private method that controls the allowed parameter combinations
        for hierarchical clustering.

        :param par_dict: clustering parameter grid.
        :type par_dict: dict
        :return: whether the parameter combination can be allowed.
        :rtype: bool
        """
        if isinstance(self.clust_method, AgglomerativeClustering):
            try:
                if par_dict['linkage'] == 'ward':
                    return par_dict['affinity'] == 'euclidean'
                else:
                    return True
            except KeyError:
                try:
                    return par_dict['affinity'] == 'euclidean'
                except KeyError:
                    return True
        else:
            return True


"""
Private functions
"""


def _return_best(val_scores):
    """
    Private function that returns indices corresponding to the best solution,
    i.e., those that minimize the validation stability scores.

    :param val_scores: list of validation scores averaged over cross-validation loops.
    :type val_scores: list
    :return: list of indices.
    :rtype: list
    """
    bidx = list(np.where(np.array(val_scores) == min([vs for vs in val_scores]))[0])
    return bidx


def _return_knownbest(val_perf, idx):
    """
    Private function that, given a stability score list and indices, returns the indices corresponding
    to the best solution.

    :param val_perf: list of validation scores averaged over cross-validation loops.
    :type val_perf: list
    :param idx: list of indices.
    :type idx: list
    :return: list of indices.
    :rtype: list
    """
    bidx = _return_best([val_perf[i] for i in idx])
    return [idx[b] for b in bidx]


def _create_result_table(out):
    """
    Private function that builds the performance result dictionary to be transformed to
    dataframe.

    :param out: grid search performance results.
    :type out: list
    :return: dictionary with results.
    :rtype: dict
    """
    dict_obj = {}
    for el in out:
        for key, val in el:
            if key in dict_obj:
                if not isinstance(dict_obj[key], list):
                    dict_obj[key] = [dict_obj[key]]
                dict_obj[key].append(val)
            else:
                dict_obj[key] = val
    return dict_obj
